"""Test adding of dispersions"""

import numpy as np
import pytest
from elli import Cauchy, Sellmeier
from elli.dispersions.base_dispersion import DispersionSum, IndexDispersionSum
from elli.dispersions.table_epsilon import TableEpsilon
from numpy.testing import assert_array_almost_equal


def test_adding_index_dispersion():
    """Test correct adding of index based dispersions"""
    index_dispersion_sum = Cauchy() + Cauchy()

    assert_array_almost_equal(
        index_dispersion_sum.get_dielectric(),
        (Cauchy().get_refractive_index() * 2) ** 2,
    )


def test_fail_on_adding_index_and_diel_dispersion():
    """Test whether the adding fails for an index based and dielectric dispersion"""

    with pytest.raises(TypeError) as sum_err:
        _ = Sellmeier() + Cauchy()

    assert (
        "Cannot add refractive index and dielectric function based dispersions."
        in str(sum_err.value)
    )


@pytest.mark.parametrize(
    "dispersions",
    [
        (DispersionSum(Sellmeier(), Sellmeier()), Cauchy()),
        (IndexDispersionSum(Cauchy(), Cauchy()), Sellmeier()),
        (
            DispersionSum(Sellmeier(), Sellmeier()),
            IndexDispersionSum(Cauchy(), Cauchy()),
        ),
        (Sellmeier(), Cauchy()),
    ],
)
def test_fail_on_adding_index_and_diel_dispersion_sum(dispersions):
    """Adding of index based and diel based dispersion sums fails"""

    with pytest.raises(TypeError):
        _ = dispersions[0] + dispersions[1]

    with pytest.raises(TypeError):
        _ = dispersions[1] + dispersions[0]


def test_fail_on_adding_index_and_diel_dispersion_as_args():
    """
    The adding of dispersion fails for an index based and dielectric dispersion
    when provided as args to DispersionSum
    """
    with pytest.raises(TypeError) as sum_err:
        DispersionSum(Sellmeier(), Cauchy())

        assert (
            "Cannot add refractive index and dielectric function based dispersions."
            in str(sum_err.value)
        )


def test_adding_of_diel_dispersions():
    """Test if dielectric dispersions are added correctly"""

    dispersion_sum = Sellmeier() + Sellmeier()

    assert isinstance(dispersion_sum, DispersionSum)
    assert len(dispersion_sum.dispersions) == 2

    for disp in dispersion_sum.dispersions:
        assert isinstance(disp, Sellmeier)

    assert_array_almost_equal(
        dispersion_sum.get_dielectric_df().values,
        2 * Sellmeier().get_dielectric_df().values,
    )


def test_flat_dispersion_sum_on_multiple_add():
    """Test whether the DispersionSum stays flat on multiple adds"""

    dispersion_sum = Sellmeier() + Sellmeier() + Sellmeier()

    assert isinstance(dispersion_sum, DispersionSum)
    assert len(dispersion_sum.dispersions) == 3

    for disp in dispersion_sum.dispersions:
        assert isinstance(disp, Sellmeier)

    assert_array_almost_equal(
        dispersion_sum.get_dielectric_df().values,
        3 * Sellmeier().get_dielectric_df().values,
    )


def test_flat_dispersion_on_adding_with_dispersion_sum():
    """
    DispersionSum is kept flat even when a mixture of
    Dispersions and DispersionSums are added
    """
    dispersion_sum = (
        Sellmeier()
        + DispersionSum(DispersionSum(Sellmeier(), Sellmeier()), Sellmeier())
        + Sellmeier()
    )

    assert isinstance(dispersion_sum, DispersionSum)
    assert len(dispersion_sum.dispersions) == 5

    for disp in dispersion_sum.dispersions:
        assert isinstance(disp, Sellmeier)

    assert_array_almost_equal(
        dispersion_sum.get_dielectric_df().values,
        5 * Sellmeier().get_dielectric_df().values,
    )


def test_multiple_dispersion_sum_args():
    """Multiple Dispersions can be provided via the DispersionSum args"""
    dispersion_sum = DispersionSum(Sellmeier(), Sellmeier(), Sellmeier())

    assert isinstance(dispersion_sum, DispersionSum)
    assert len(dispersion_sum.dispersions) == 3

    for disp in dispersion_sum.dispersions:
        assert isinstance(disp, Sellmeier)

    assert_array_almost_equal(
        dispersion_sum.get_dielectric_df().values,
        3 * Sellmeier().get_dielectric_df().values,
    )


def test_nested_dispersion_sum_args():
    """
    DispersionSum is kept flat even for nested
    DispersionSum args
    """
    dispersion_sum = DispersionSum(
        DispersionSum(DispersionSum(Sellmeier(), Sellmeier()), Sellmeier()), Sellmeier()
    )

    assert isinstance(dispersion_sum, DispersionSum)
    assert len(dispersion_sum.dispersions) == 4

    for disp in dispersion_sum.dispersions:
        assert isinstance(disp, Sellmeier)

    assert_array_almost_equal(
        dispersion_sum.get_dielectric_df().values,
        4 * Sellmeier().get_dielectric_df().values,
    )


def test_flattening_of_dispersion_sum_args():
    """When a DispersionSum is passed via a DisperionSum arg it is flattened."""
    dispersion_sum = DispersionSum(Sellmeier(), DispersionSum(Sellmeier(), Sellmeier()))

    assert isinstance(dispersion_sum, DispersionSum)
    assert len(dispersion_sum.dispersions) == 3

    for disp in dispersion_sum.dispersions:
        assert isinstance(disp, Sellmeier)

    assert_array_almost_equal(
        dispersion_sum.get_dielectric_df().values,
        3 * Sellmeier().get_dielectric_df().values,
    )


def test_adding_of_float_to_tabular_epsilon_dispersion():
    """Tests correct adding of a float value to tabular dispersions"""

    table_eps = TableEpsilon(lbda=np.linspace(200, 1000, 801), epsilon=np.ones(801)) + 1

    assert_array_almost_equal(
        table_eps.get_dielectric(),
        np.ones(801) * 2,
    )


def test_adding_of_dispersion_function_and_table():
    """Dispersion function and table can be added"""

    table = TableEpsilon(lbda=np.linspace(200, 1000, 801), epsilon=np.ones(801))
    sellmeier = Sellmeier().add(1, 0.0001)

    assert_array_almost_equal(
        (table + sellmeier).get_dielectric(),
        table.get_dielectric() + sellmeier.get_dielectric(),
    )


def test_fail_on_adding_tabular_dispersions():
    """Adding two tabular dispersions should return an NotImplementedError"""

    with pytest.raises(NotImplementedError) as not_impl_err:
        _ = TableEpsilon(
            lbda=np.linspace(200, 1000, 801), epsilon=np.ones(801)
        ) + TableEpsilon(lbda=np.linspace(200, 1000, 801), epsilon=np.ones(801))

    assert (
        str(not_impl_err.value) == "Adding of tabular dispersions is not yet supported"
    )



"""A formula dispersion to parse dispersion values from a formula string."""

from typing import Dict, List, Optional

import numpy as np
import numpy.typing as npt

from elli.units import ureg
from elli.dispersions.base_dispersion import BaseDispersion, Dispersion, IndexDispersion
from elli.formula_parser.parser import parse_formula, FormulaTransformer


class FormulaParser(BaseDispersion):
    r"""A formula dispersion"""

    @property
    def single_params_template(self) -> dict:
        return self.f_single_params

    @property
    def rep_params_template(self) -> dict:
        return self.f_rep_params

    def __init__(
        self,
        formula: str,
        wavelength_axis_name: str,
        single_params: Dict[str, float],
        rep_params: Dict[str, npt.ArrayLike],
        unit: Optional[str] = None,
    ):
        self.f_single_params: Dict[str, float] = single_params
        self.f_axis_name: str = wavelength_axis_name
        rep_params_len: Optional[int] = None
        rep_params_sets: List[Dict[str, float]] = []

        for key, values in rep_params.items():
            if not isinstance(values, (np.ndarray, list)):
                raise ValueError(
                    "Repeated parameters must be given as dict of lists or numpy arrays"
                )
            for i, value in enumerate(values):
                if i >= len(rep_params_sets):
                    rep_params_sets.append({})
                rep_params_sets[i][key] = value

            if rep_params_len is None:
                rep_params_len = len(values)
                continue
            if len(values) != rep_params_len:
                raise ValueError(
                    f"All repeated parameters must have the same length."
                    f"Found {values} with length {len(values)}, "
                    f"but previous length was {rep_params_len}."
                )

        self.f_rep_params = {}
        if rep_params_sets:
            self.f_rep_params = rep_params_sets[0]
        super().__init__()

        for rep_params_set in rep_params_sets:
            self.add(**rep_params_set)

        self.rep_params_dl = {}
        if self.rep_params:
            self.rep_params_dl = {
                k: np.array([dic[k] for dic in self.rep_params])
                for k in self.rep_params[0]
            }

        self.formula = formula

        self._check_repr()

        self._dispersion_function = self.__dispersion_function
        if unit is not None:
            self._set_unit_conversion(unit)

    def _set_unit_conversion(self, unit: str):
        quantity = ureg(unit)
        if quantity.check("[length]"):
            scaling = ureg("nm").to(unit).magnitude
            self._dispersion_function = lambda lbda: self.__dispersion_function(
                scaling * lbda
            )
            return

        if quantity.check("[energy]"):
            scaling = ureg("nm").to(unit).magnitude
            self._dispersion_function = lambda lbda: self.__dispersion_function(
                scaling / lbda
            )
            return

        raise ValueError(f"Unsupported unit `{unit}`.")

    def _check_repr(self):
        representation = parse_formula(self.formula).data

        if isinstance(self, FormulaIndex) and not representation == "n":
            raise ValueError(
                f"Representation `{representation}` not supported by FormulaIndex"
            )

        if isinstance(self, Formula) and not representation == "eps":
            raise ValueError(
                f"Representation `{representation}` not supported by Formula"
            )

    def __dispersion_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        return FormulaTransformer(
            x_axis_name=self.f_axis_name,
            x_axis_values=lbda,
            single_params=self.single_params,
            repeated_params=self.rep_params_dl,
        ).transform(parse_formula(self.formula))[1]


class Formula(Dispersion, FormulaParser):
    r"A formula dispersion"

    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        return self._dispersion_function(lbda)


class FormulaIndex(IndexDispersion, FormulaParser):
    r"""A formula dispersion in refractive index formulation"""

    def refractive_index(self, lbda: npt.ArrayLike) -> npt.NDArray:
        return self._dispersion_function(lbda)



"""Tests for reading woollam"""

import pytest
from fixtures import datadir  # pylint: disable=unused-import
import elli


# pylint: disable=redefined-outer-name
def test_reading_of_psi_delta_woollam(datadir):
    """Psi/delta Spectraray file is read w/o errors"""
    data_wvase = elli.read_woollam_psi_delta(datadir / "wvase_example.dat")
    data_cease = elli.read_woollam_psi_delta(datadir / "complete_ease_example.dat")

    assert data_wvase.shape == (542, 2)
    assert data_cease.shape == (3263, 2)


# pylint: disable=redefined-outer-name
def test_reading_and_conv_to_woollam(datadir):
    """Rho values are read from Psi / Delta file"""
    data_wvase = elli.read_woollam_rho(datadir / "wvase_example.dat")
    data_cease = elli.read_woollam_rho(datadir / "complete_ease_example.dat")

    assert data_wvase.shape == (542,)
    assert data_cease.shape == (3263,)


# pylint: disable=redefined-outer-name
def test_raises_not_implemented_for_tan_cos_format(datadir):
    """Raises error when a wvase Tan(Psi)/Cos(Delta) file is presented"""

    with pytest.raises(NotImplementedError):
        elli.read_woollam_psi_delta(datadir / "wvase_trig.dat")



"""Decorator functions for convenient fitting of Mueller matrices"""

# Encoding: utf-8
from typing import Callable

import numpy.typing as npt
import pandas as pd

try:
    from lmfit import Parameters, minimize
    from lmfit.minimizer import MinimizerResult
    import plotly.graph_objects as go
    from IPython.display import display
    from ipywidgets import widgets
except ImportError as e:
    raise ImportError(
        "This module requires lmfit, ipywidgets, plotly and ipython.\n"
        "Try installing this package with the additional fitting requirement, "
        "i.e. pip install pyElli[fitting]"
    ) from e

from ..plot.mueller_matrix import plot_mmatrix
from ..result import Result
from .decorator import FitDecorator, is_in_notebook
from .params_hist import ParamsHist


def mmatrix_to_dataframe(
    exp_df: pd.DataFrame, mueller_matrix: npt.NDArray, identifier: str = None
) -> pd.DataFrame:
    """Reshape a numpy 4x4 array containing mueller matrix elements
    to a dataframe with columns Mxy. The index labels for each column
    are taken from the provided exp_df.

    Args:
        exp_df (pd.DataFrame): The experimental dataframe providing the index and columns for
                               the newly generated dataframe.
        mueller_matrix (npt.NDArray): Data to be reshaped into a dataframe
        identifier (str, optional):
            An identifier to append to each column name, in the form
            Mxy_<identifier>, where Mxy is the old column name.
            Defaults to None.

    Returns:
        pd.DataFrame: Contains the data from mueller_matrix in the shape of exp_df
    """
    if identifier is not None:
        columns = [f"{c}_{identifier}" for c in exp_df.columns]
    else:
        columns = exp_df.columns
    mueller_df = pd.DataFrame(index=exp_df.index, columns=columns, dtype="float64")
    mueller_df.values[:] = mueller_matrix.reshape(-1, 16)

    return mueller_df


class FitMuellerMatrix(FitDecorator):
    """A class to fit mueller matrices to experimental data"""

    def update_selection(self, _: dict = None) -> None:
        """Update plot after selection of displayed data

        Args:
            _ (dict, optional): No function. Just for compliance with ABC.
        """
        with self.fig.batch_update():
            if self.show_residual:
                model_df = (
                    mmatrix_to_dataframe(
                        self.exp_mm,
                        self.model(
                            self.exp_mm.index.values, self.params
                        ).mueller_matrix,
                    )
                    - self.exp_mm
                )
            else:
                model_df = mmatrix_to_dataframe(
                    self.exp_mm,
                    self.model(self.exp_mm.index.values, self.params).mueller_matrix,
                )

            for i, melem in enumerate(model_df):
                self.fig.data[2 * i + 1].y = model_df[melem]
                if self.show_residual:
                    self.fig.data[2 * i + 1].name = f"{melem} Residual"
                else:
                    self.fig.data[2 * i + 1].name = f"{melem} theory"

    def update_residual(self, change: dict) -> None:
        self.show_residual = change.new
        self.update_selection()

    def create_widgets(self) -> None:
        """Create ipywidgets for parameter estimation"""
        self.param_widgets = {}
        checkboxes = []
        for param in self.params.valuesdict():
            curr_checkbox = widgets.Checkbox(
                value=self.params[param].vary,
                description_tooltip=param,
                layout=widgets.Layout(width="15px"),
                indent=False,
            )
            curr_checkbox.observe(self.set_vary_param, names="value")
            curr_widget = widgets.BoundedFloatText(
                self.params[param],
                min=self.params[param].min,
                max=self.params[param].max,
                description=param,
                continuous_update=False,
            )
            curr_widget.observe(self.update_params, names=("value", "owner"))
            self.param_widgets[param] = curr_widget
            checkboxes.append(curr_checkbox)

        fit_button = widgets.Button(description="Fit")
        fit_button.on_click(lambda _: self.fit_button_clicked())
        button_list = [fit_button]

        if isinstance(self.params, ParamsHist):
            undo_button = widgets.Button(description="Undo")
            undo_button.on_click(self.re_undo_button_clicked)
            redo_button = widgets.Button(description="Redo")
            redo_button.on_click(self.re_undo_button_clicked)
            init_button = widgets.Button(description="Set initial")
            init_button.on_click(lambda _: self.reset_to_init_params())

            button_list.append(undo_button)
            button_list.append(redo_button)
            button_list.append(init_button)

        residual_checkbox = widgets.Checkbox(
            value=self.show_residual, description="Display Residual"
        )
        residual_checkbox.observe(self.update_residual, names="value")

        combo_widget = [
            widgets.HBox([v, w])
            for v, w in zip(list(self.param_widgets.values()), checkboxes)
        ]
        if is_in_notebook():
            display(
                widgets.VBox(
                    [
                        widgets.HBox(
                            combo_widget + button_list + [residual_checkbox],
                            layout=widgets.Layout(
                                width="100%",
                                display="inline-flex",
                                flex_flow="row wrap",
                            ),
                        ),
                        self.fig,
                    ]
                )
            )

    def fit_function(
        self, params: Parameters, lbda: npt.NDArray, mueller_matrix: pd.DataFrame
    ) -> npt.NDArray:
        """The fit function to minimize the fitting problem

        Args:
            params (Parameters): The lmfit fitting Parameters to construct the simulation
            lbda (npt.NDArray): Wavelengths in nm
            mueller_matrix (pd.DataFrame): The experimental data to compare to the fitted model

        Returns:
            npt.NDArray: Residual between the calculation
                with current parameters and experimental data
        """
        return (
            mueller_matrix.values.reshape(-1, 4, 4)
            - self.model(lbda, params).mueller_matrix
        )

    def fit(self, method: str = "leastsq") -> MinimizerResult:
        """Execute lmfit with the current fitting parameters

        Args:
            method (str, optional): The fitting method to use.
                Any method supported by scipys curve_fit is allowed.
                Defaults to 'leastsq'.

        Returns:
            Result: The fitting result
        """
        res = minimize(
            self.fit_function,
            self.params,
            args=(self.exp_mm.index.values, self.exp_mm),
            method=method,
        )

        self.fitted_params = res.params
        return res

    def get_model_data(
        self, params: Parameters = None, append_exp_data=False
    ) -> pd.DataFrame:
        """Gets the data from the provided model with the provided parameters.
        If no parameters are provided, the fitted parameters are used
        (which default to the initial parameters if no fit has been triggered).

        Args:
            params (Parameters, optional):
                The parameters to calculate the model with.
                If not provided, the fitted parameters are used.
                Defaults to None.
            append_exp_data (bool, optional):
                Appends the experimental data if set to True.
                Defaults to False.

        Returns:
            pd.DataFrame: The model results
        """
        if params is None:
            fit_result = self.model(self.exp_mm.index.values, self.fitted_params)
            desc = "fit"
        else:
            fit_result = self.model(self.exp_mm.index.values, params)
            desc = "model"

        if append_exp_data:
            return pd.concat(
                [
                    self.exp_mm,
                    mmatrix_to_dataframe(
                        self.exp_mm, fit_result.mueller_matrix, identifier=desc
                    ),
                ],
                axis=1,
            )

        return mmatrix_to_dataframe(
            self.exp_mm, fit_result.mueller_matrix, identifier=desc
        )

    def plot(self, **kwargs) -> go.Figure:
        """Plot the fit results

        Args:
            **display_single (bool):
                Returns a figure containing a single graph, if set to true.
                Returns a grid of figures otherwise.
            **sharex (bool): Ties the zoom of the x-axes together for grid view.
            **full_scale (bool): Sets the y-axis scale to [-1, 1] if set to True.

        Returns:
            go.Figure: The figure containing the data
        """
        fit_result = mmatrix_to_dataframe(
            self.exp_mm,
            self.model(self.exp_mm.index.values, self.fitted_params).mueller_matrix,
        )

        return plot_mmatrix(
            [self.exp_mm, fit_result],
            single=(
                self.display_single
                if kwargs.get("display_single") is None
                else kwargs.get("display_single")
            ),
            sharex=(
                self.sharex if kwargs.get("sharex") is None else kwargs.get("sharex")
            ),
            full_scale=(
                self.full_scale
                if kwargs.get("full_scale") is None
                else kwargs.get("full_scale")
            ),
        )

    def plot_residual(self, **kwargs) -> go.Figure:
        """Plots the residual between the fit and the experimental data

        Args:
            **display_single (bool):
                Returns a figure containing a single graph, if set to true.
                Returns a grid of figures otherwise.
            **sharex (bool): Ties the zoom of the x-axes together for grid view.
            **full_scale (bool): Sets the y-axis scale to [-1, 1] if set to True.

        Returns:
            go.Figure: The figure containing the data
        """
        fit_result = mmatrix_to_dataframe(
            self.exp_mm,
            self.model(self.exp_mm.index.values, self.fitted_params).mueller_matrix,
        )

        return plot_mmatrix(
            [fit_result - self.exp_mm],
            single=(
                self.display_single
                if kwargs.get("display_single") is None
                else kwargs.get("display_single")
            ),
            sharex=(
                self.sharex if kwargs.get("sharex") is None else kwargs.get("sharex")
            ),
            full_scale=(
                self.full_scale
                if kwargs.get("full_scale") is None
                else kwargs.get("full_scale")
            ),
        )

    def __init__(
        self,
        exp_mm: pd.DataFrame,
        params: Parameters,
        model: Callable[[npt.NDArray, Parameters], Result],
        **kwargs,
    ) -> None:
        """Intialize the mueller matrix fitting class

        Args:
            exp_mm (pd.DataFrame): The dataframe containing an experimental mueller matrix.
                                   It should contain 16 columns with labels Mxy,
                                   where xy are the matrix positions.
            params (Parameters): Fitting start parameters
            model (Callable[[npt.NDArray, Parameters], Result]):
                A function taking wavelengths as first parameter
                and fitting parameters as second,
                which returns a pyEllis Result object.
                This function contains the actual model which should be fitted

            **display_single (bool):
                Returns a figure containing a single graph, if set to true.
                Returns a grid of figures otherwise.
            **sharex (bool): Ties the zoom of the x-axes together for grid view.
            **full_scale (bool): Sets the y-axis scale to [-1, 1] if set to True.
        """
        super().__init__()
        self.exp_mm = exp_mm
        self.params = params
        self.fitted_params = params.copy()
        self.initial_params = params.copy()
        self.model = model
        self.param_widgets = {}
        self.show_residual = False

        mm_kwargs = ("display_single", "sharex", "full_scale")
        for kwarg in mm_kwargs:
            if kwarg in kwargs:
                setattr(self, kwarg, kwargs[kwarg])
                del kwargs[kwarg]
        self.fit_kwargs = kwargs

        model_df = mmatrix_to_dataframe(
            exp_mm, model(exp_mm.index.values, params).mueller_matrix
        )
        self.fig = plot_mmatrix(
            [exp_mm, model_df],
            single=self.display_single,
            sharex=self.sharex,
            full_scale=self.full_scale,
        )

        self.create_widgets()


def fit_mueller_matrix(
    exp_mm: pd.DataFrame, params: Parameters, **kwargs
) -> Callable[[npt.NDArray, Parameters], Result]:
    """A parameters decorator for fitting mueller matrices. Displays an ipywidget float box for
    each fitting parameter and an interactive plot to estimate parameters.

    Args:
        exp_mm (pd.DataFrame): The dataframe containing an experimental mueller matrix.
                               It should contain 16 columns with labels Mxy,
                               where xy are the matrix positions.
        params (Parameters): Fitting start parameters
        **display_single (bool):
            Returns a figure containing a single graph, if set to true.
            Returns a grid of figures otherwise.
        **sharex (bool): Ties the zoom of the x-axes together for grid view.
        **full_scale (bool): Sets the y-axis scale to [-1, 1] if set to True.

    Returns:
        Callable[[npt.NDArray, Parameters], Result]:
            A function taking wavelengths as first parameter and fitting parameters as second,
            which returns a pyEllis Result object.
            This function contains the actual model which should be fitted and is automatically
            provided when used as a decorator.
    """
    return lambda model: FitMuellerMatrix(exp_mm, params, model, **kwargs)



# Encoding: utf-8
import warnings

import numpy as np
from numpy.lib.scimath import arcsin, sqrt

from .result import Result
from .solver import Solver


class Solver2x2(Solver):
    """
    Solver class to evaluate Experiment objects.
    Simple but fast 2x2 transfer matrix method.
    Cannot handle anisotropy or anything fancy,
    thus Jonas and Mueller matrices cannot be calculated (respective functions return None).
    """

    def list_snell(self, n_list):
        angles = arcsin(n_list[0] * np.sin(np.deg2rad(self.theta_i)) / n_list)

        angles[0] = np.where(
            np.invert(Solver2x2.is_forward_angle(n_list[0], angles[0])),
            np.pi - angles[0],
            angles[0],
        )
        angles[-1] = np.where(
            np.invert(Solver2x2.is_forward_angle(n_list[-1], angles[-1])),
            np.pi - angles[-1],
            angles[-1],
        )

        return angles

    def calculate(self) -> Result:
        """Calculates the transfer matrix for the given material stack"""
        if len(self.permittivity_profile) > 2:
            d, eps = list(zip(*self.permittivity_profile[1:-1]))
            d_list = np.array(d)
            n_list = sqrt(
                np.vstack(
                    [
                        self.permittivity_profile[0][1][:, 0, 0],
                        np.array(eps)[..., 0, 0],
                        self.permittivity_profile[-1][1][:, 0, 0],
                    ]
                )
            )
        else:
            d_list = np.array([])
            n_list = sqrt(
                np.vstack(
                    [
                        self.permittivity_profile[0][1][:, 0, 0],
                        self.permittivity_profile[-1][1][:, 0, 0],
                    ]
                )
            )

        for layer in n_list:
            if np.any(np.logical_and(layer.real > 0, layer.imag < 0)):
                warnings.warn(
                    """Solver2x2 can't handle active media (n > 0 and k < 0).
                    Check if all materials are defined correctly or switch to Solver4x4 instead."""
                )
            elif np.any(np.logical_and(layer.real < 0, layer.imag > 0)):
                warnings.warn(
                    """Solver2x2 can't handle media with n < 0 and k > 0.
                    Check if all materials are defined correctly or switch to Solver4x4 instead."""
                )

        num_layers = n_list.shape[0]
        th_list = self.list_snell(n_list)
        kz_list = 2 * np.pi * n_list * np.cos(th_list) / self.lbda

        delta = kz_list[1:-1] * (d_list if n_list.ndim == 1 else d_list[:, None])

        esum = "ij...,jk...->ik..."
        ones = np.repeat(1, n_list.shape[1]) if n_list.ndim > 1 else 1

        rs, rp, ts, tp = Solver2x2.fresnel(n_list[0], n_list[1], th_list[0], th_list[1])
        Ms = np.array([[ones, rs], [rs, ones]], dtype=complex) / ts
        Mp = np.array([[ones, rp], [rp, ones]], dtype=complex) / tp

        for i in range(1, num_layers - 1):
            rs, rp, ts, tp = Solver2x2.fresnel(
                n_list[i], n_list[i + 1], th_list[i], th_list[i + 1]
            )
            em = np.exp(-1j * delta[i - 1])
            ep = np.exp(1j * delta[i - 1])

            Ms = np.einsum(
                esum, Ms, np.array([[em, rs * em], [rs * ep, ep]], dtype=complex) / ts
            )
            Mp = np.einsum(
                esum, Mp, np.array([[em, rp * em], [rp * ep, ep]], dtype=complex) / tp
            )

        rtots = Ms[1, 0] / Ms[0, 0]
        ttots = 1 / Ms[0, 0]
        rtotp = Mp[1, 0] / Mp[0, 0]
        ttotp = 1 / Mp[0, 0]

        zeros = np.repeat(0, n_list.shape[1]) if n_list.ndim > 1 else 1

        jones_matrix_r = np.moveaxis(np.array([[rtotp, zeros], [zeros, rtots]]), 2, 0)
        jones_matrix_t = np.moveaxis(np.array([[ttotp, zeros], [zeros, ttots]]), 2, 0)

        # TODO: Test if p and s correction formulas are needed.
        power_correction = ((n_list[-1] * np.cos(th_list[-1])).real) / (
            n_list[0] * np.cos(th_list[0])
        ).real

        return Result(self.experiment, jones_matrix_r, jones_matrix_t, power_correction)

    @staticmethod
    def fresnel(n_i, n_t, th_i, th_t):
        r"""Calculate fresnel coefficients at the interface of two materials

        Args:
            n_i: Refractive index of the material of the incident wave
            n_t: Refractive index of the material of the transmitted wave
            thi_i: Incident angle of the incident wave
            thi_t: Refracted angle of the transmitted wave

        Returns:
            r_s: s-polarized reflection coefficient
            r_p: p-polarized reflection coefficient
            t_s: s-polarized transmission coefficient
            t_p: p-polarized transmission coefficient
        """
        cos_i = np.cos(th_i)
        cos_t = np.cos(th_t)

        r_s = (n_i * cos_i - n_t * cos_t) / (n_i * cos_i + n_t * cos_t)
        r_p = (n_t * cos_i - n_i * cos_t) / (n_t * cos_i + n_i * cos_t)
        t_s = 2 * n_i * cos_i / (n_i * cos_i + n_t * cos_t)
        t_p = 2 * n_i * cos_i / (n_t * cos_i + n_i * cos_t)

        return r_s, r_p, t_s, t_p

    @staticmethod
    def is_forward_angle(n, theta):
        ncostheta = n * np.cos(theta)
        return np.where(
            abs(ncostheta.imag) > 1e-10, ncostheta.imag > 0, ncostheta.real > 0
        )



# Encoding: utf-8
"""The Materials submodule provides the classes to build
complex materials from :doc:`dispersion models<dispersions>`.

The simplest provided material is the
:class:`IsotropicMaterial<elli.materials.IsotropicMaterial>`.
It takes only one Dispersion which is used for all three axis of the material.
It can also be created by calling dispersion.get_mat().

For Crystals with two or three different dispersions, there are the
:class:`UniaxialMaterial<elli.materials.UniaxialMaterial>` and
:class:`UniaxialMaterial<elli.materials.UniaxialMaterial>` respectively.

The respective materials can be rotated to achieve different crystal orientations.
A good visualization for these rotations can be found in Figure 6.10 of Fujiwara's
book 'Spectroscopic Ellipsometry' [1]_.

Additionally two materials can be combined via various :ref:'Effective medium approximations',
to create mixtures or account for interface roughness.

.. rubric:: References

.. [1] H. Fujiwara,
   Spectroscopic Ellipsometry,
   Principles and Applications,
   Chichester, UK, John Wiley & Sons, Ltd (2007).
   https://doi.org/10.1002/9780470060193
"""

from abc import ABC, abstractmethod

import numpy as np
import numpy.typing as npt
from numpy.lib.scimath import sqrt, power

from .dispersions.base_dispersion import BaseDispersion


class Material(ABC):
    """Base class for materials (abstract class)."""

    @abstractmethod
    def get_tensor(self, lbda: npt.ArrayLike) -> npt.NDArray:
        """Gets the permittivity tensor of the material for wavelength 'lbda'.

        Args:
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).

        Returns:
            npt.NDArray: Permittivity tensor.
        """

    def get_refractive_index(self, lbda: npt.ArrayLike) -> npt.NDArray:
        """Gets the refractive index tensor for wavelength 'lbda'.

        Args:
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).

        Returns:
            npt.NDArray: Refractive index tensor.
        """
        return sqrt(self.get_tensor(lbda))


class SingleMaterial(Material):
    """Base class for non-mixed materials (abstract class)."""

    dispersion_x = None
    dispersion_y = None
    dispersion_z = None
    rotated = False
    rotation_matrix = None

    @abstractmethod
    def set_dispersion(self) -> None:
        """Sets dispersion relation of the material."""

    def set_rotation(self, r: npt.NDArray) -> None:
        """Sets rotation of the Material.

        Args:
            r (npt.NDArray):
                rotation matrix (from :func:`rotation_euler<elli.utils.rotation_euler>` or others)
        """
        self.rotated = True
        self.rotation_matrix = r

    def get_tensor(self, lbda: npt.ArrayLike) -> npt.NDArray:
        """Gets the permittivity tensor of the material for wavelength 'lbda'.

        Args:
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).

        Returns:
            npt.NDArray: Permittivity tensor.
        """
        # Check for shape of lbda
        shape = np.shape(lbda)
        if shape == ():
            length = 1
        else:
            length = shape[0]

        # create empty tensor
        epsilon = np.zeros((length, 3, 3), dtype=np.complex128)

        # get get dielectric functions from dispersion
        epsilon[:, 0, 0] = self.dispersion_x.get_dielectric(lbda)
        epsilon[:, 1, 1] = self.dispersion_y.get_dielectric(lbda)
        epsilon[:, 2, 2] = self.dispersion_z.get_dielectric(lbda)

        if self.rotated:
            epsilon = self.rotation_matrix @ epsilon @ self.rotation_matrix.T

        return epsilon


class IsotropicMaterial(SingleMaterial):
    """Isotropic material."""

    def __init__(self, dispersion: BaseDispersion) -> None:
        """Creates isotropic material with a dispersion.

        Args:
            dispersion (Dispersion): Dispersion relation of all three crystal directions.
        """
        self.set_dispersion(dispersion)

    # pylint: disable=arguments-differ
    def set_dispersion(self, dispersion: BaseDispersion) -> None:
        """Sets dispersion relation of the isotropic material.

        Args:
            dispersion (Dispersion): Dispersion relation of all three crystal directions.
        """
        if not isinstance(dispersion, BaseDispersion):
            raise TypeError(
                f"Expected dispersion to be an Dispersion object but found type {type(dispersion)}."
            )

        self.dispersion_x = dispersion
        self.dispersion_y = dispersion
        self.dispersion_z = dispersion


class UniaxialMaterial(SingleMaterial):
    """Uniaxial material."""

    def __init__(
        self, dispersion_o: BaseDispersion, dispersion_e: BaseDispersion
    ) -> None:
        """Creates a uniaxial material with two dispersions.

        Args:
            dispersion_o (Dispersion):
                Dispersion relation for ordinary crystal axes (x and y direction).
            dispersion_e (Dispersion):
                Dispersion relation for extraordinary crystal axis (z direction).
        """
        self.set_dispersion(dispersion_o, dispersion_e)

    # pylint: disable=arguments-differ
    def set_dispersion(
        self, dispersion_o: BaseDispersion, dispersion_e: BaseDispersion
    ) -> None:
        """Sets dispersion relations of the uniaxial material.

        Args:
            dispersion_o (Dispersion):
                Dispersion relation for ordinary crystal axes (x and y direction).
            dispersion_e (Dispersion):
                Dispersion relation for extraordinary crystal axis (z direction).
        """
        for dispersion in [dispersion_o, dispersion_e]:
            if not isinstance(dispersion, BaseDispersion):
                raise TypeError(
                    f"Expected dispersion to be an Dispersion object but found type {type(dispersion)}."
                )

        self.dispersion_x = dispersion_o
        self.dispersion_y = dispersion_o
        self.dispersion_z = dispersion_e


class BiaxialMaterial(SingleMaterial):
    """Biaxial material."""

    def __init__(
        self,
        dispersion_x: BaseDispersion,
        dispersion_y: BaseDispersion,
        dispersion_z: BaseDispersion,
    ) -> None:
        """Creates a biaxial material with three dispersions.

        Args:
            dispersion_x (Dispersion): Dispersion relation for x crystal axes.
            dispersion_y (Dispersion): Dispersion relation for y crystal axes.
            dispersion_z (Dispersion): Dispersion relation for z crystal axes.
        """
        self.set_dispersion(dispersion_x, dispersion_y, dispersion_z)

    # pylint: disable=arguments-differ
    def set_dispersion(
        self,
        dispersion_x: BaseDispersion,
        dispersion_y: BaseDispersion,
        dispersion_z: BaseDispersion,
    ) -> None:
        """Sets dispersion relations of the biaxial material.

        Args:
            dispersion_x (Dispersion): Dispersion relation for x crystal axes.
            dispersion_y (Dispersion): Dispersion relation for y crystal axes.
            dispersion_z (Dispersion): Dispersion relation for z crystal axes.
        """
        for dispersion in [dispersion_x, dispersion_y, dispersion_z]:
            if not isinstance(dispersion, BaseDispersion):
                raise TypeError(
                    f"Expected dispersion to be an Dispersion object but found type {type(dispersion)}."
                )

        self.dispersion_x = dispersion_x
        self.dispersion_y = dispersion_y
        self.dispersion_z = dispersion_z


class MixtureMaterial(Material):
    """Abstract Class for mixed materials."""

    host_material = None
    guest_material = None
    fraction = None

    def __init__(
        self, host_material: Material, guest_material: Material, fraction: float
    ) -> None:
        """Creates a material mixture from two materials.

        Args:
            host_material (Material): Host Material.
            guest_material (Material): Material incorporated in the host.
            fraction (float): Fraction of the guest material (Range 0 - 1).
        """
        self.set_constituents(host_material, guest_material)
        self.set_fraction(fraction)

    def set_constituents(
        self, host_material: Material, guest_material: Material
    ) -> None:
        """Sets Materials in the mixture.

        Args:
            host_material (Material): Host Material.
            guest_material (Material): Material incorporated in the host.
        """
        if not isinstance(host_material, Material):
            raise TypeError(
                f"Expected material to be an Material object but found type {type(host_material)}."
            )
        if not isinstance(guest_material, Material):
            raise TypeError(
                f"Expected material to be an Material object but found type {type(guest_material)}."
            )

        self.host_material = host_material
        self.guest_material = guest_material

    def set_fraction(self, fraction: float) -> None:
        """Sets fraction and checks if fraction is in range from 0 to 1.

        Args:
            fraction (float): Fraction of the guest material (Range 0 - 1).
        """
        if not 0 <= fraction <= 1:
            raise ValueError("Fraction is not in range from 0 to 1")

        self.fraction = fraction

    @abstractmethod
    def get_tensor_fraction(self, lbda: npt.ArrayLike, fraction: float) -> npt.NDArray:
        """Gets the permittivity tensor of the material for wavelength 'lbda',
        while overwriting the set fraction. Used in VaryingMixtureLayers.

        Args:
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).
            fraction (float): Fraction of the guest material used for evaluation. (Range 0 - 1).

        Returns:
            npt.NDArray: Permittivity tensor.
        """

    def get_tensor(self, lbda: npt.ArrayLike) -> npt.NDArray:
        """Gets the permittivity tensor of the material for wavelength 'lbda'.

        Args:
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).

        Returns:
            npt.NDArray: Permittivity tensor.
        """
        return self.get_tensor_fraction(lbda, self.fraction)


class VCAMaterial(MixtureMaterial):
    r"""Mixture Material approximated with a simple virtual crystal like average.

    .. math::
       \varepsilon_\text{eff} = (1 - f) \varepsilon_h + f \varepsilon_g

    where:

    * :math:`\varepsilon_\text{eff}` is the effective permittivity of host/mixture material,
    * :math:`\varepsilon_h` is the permittivity of the host mixture material,
    * :math:`\varepsilon_g` is the permittivity of the guest mixture material and
    * :math:`f` is the volume fraction of the guest in the host material.
    """

    def get_tensor_fraction(self, lbda: npt.ArrayLike, fraction: float) -> npt.NDArray:
        """Gets the permittivity tensor of the marterial for wavelength 'lbda',
        while overwriting the set fraction.

        Args:
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).
            fraction (float): Fraction of the guest material used for evaluation. (Range 0 - 1).

        Returns:
            npt.NDArray: Permittivity tensor.
        """
        epsilon = (
            self.host_material.get_tensor(lbda) * (1 - fraction)
            + self.guest_material.get_tensor(lbda) * fraction
        )
        return epsilon


class LooyengaEMA(MixtureMaterial):
    r"""Mixture Material approximated with Looyenga's formula.
    Valid for materials with small contrast.

    .. math::
       \varepsilon_\text{eff} = ((1 - f) \varepsilon_h^{1/3} + f \varepsilon_g^{1/3})^3

    where:

    * :math:`\varepsilon_\text{eff}` is the effective permittivity of host/mixture material,
    * :math:`\varepsilon_h` is the permittivity of the host mixture material,
    * :math:`\varepsilon_g` is the permittivity of the guest mixture material and
    * :math:`f` is the volume fraction of the guest in the host material.

    References:
        Looyenga, H. (1965). Physica, 31(3), 401–406.
    """

    def get_tensor_fraction(self, lbda: npt.ArrayLike, fraction: float) -> npt.NDArray:
        """Gets the permittivity tensor of the material for wavelength 'lbda',
        while overwriting the set fraction.

        Args:
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).
            fraction (float): Fraction of the guest material used for evaluation. (Range 0 - 1).

        Returns:
            npt.NDArray: Permittivity tensor.
        """
        epsilon = (
            self.host_material.get_tensor(lbda) ** (1 / 3) * (1 - fraction)
            + self.guest_material.get_tensor(lbda) ** (1 / 3) * fraction
        ) ** 3
        return epsilon


class MaxwellGarnettEMA(MixtureMaterial):
    r"""Mixture Material approximated with the Maxwell Garnett formula.
    It is valid for spherical inclusions with small volume fraction.

    .. math::
       \varepsilon_\text{eff} = \varepsilon_h \frac{2f(\varepsilon_g - \varepsilon_h)
       + \varepsilon_g + 2\varepsilon_h}
       {2\varepsilon_h + \varepsilon_g - f(\varepsilon_g - \varepsilon_h)}

    where:

    * :math:`\varepsilon_\text{eff}` is the effective permittivity of host/mixture material,
    * :math:`\varepsilon_h` is the permittivity of the host mixture material,
    * :math:`\varepsilon_g` is the permittivity of the guest mixture material and
    * :math:`f` is the volume fraction of the guest in the host material.
    """

    def get_tensor_fraction(self, lbda: npt.ArrayLike, fraction: float) -> npt.NDArray:
        """Gets the permittivity tensor of the material for wavelength 'lbda',
        while overwriting the set fraction.

        Args:
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).
            fraction (float): Fraction of the guest material used for evaluation. (Range 0 - 1).

        Returns:
            npt.NDArray: Permittivity tensor.
        """
        e_h = self.host_material.get_tensor(lbda)
        e_g = self.guest_material.get_tensor(lbda)

        # Catch calculation warnings
        old_settings = np.geterr()
        np.seterr(invalid="ignore")

        maxwell_garnett = (
            e_h
            * (2 * fraction * (e_g - e_h) + e_g + 2 * e_h)
            / (2 * e_h + e_g - fraction * (e_g - e_h))
        )

        # Reset numpy settings
        np.seterr(**old_settings)

        epsilon = np.where(np.logical_and(e_h == 0, e_g == 0), e_h, maxwell_garnett)
        return epsilon


class BruggemanEMA(MixtureMaterial):
    r"""Mixture Material approximated with the Bruggeman formula
    for isotropic spherical inclusions.

    Returns one of the two analytical solutions to this quadratic equation:

    .. math::
       2 \varepsilon_\text{eff}^2 +
       [(3f - 2) \varepsilon_a + (1 - 3f)\varepsilon_b] \varepsilon_\text{eff}
       - \varepsilon_a \cdot \varepsilon_b = 0

    where :math:`\varepsilon_\text{eff}` is the effective permittivity of host/mixture material,
    :math:`\varepsilon_a` is the permittivity of the first mixture material,
    :math:`\varepsilon_b` is the permittivity of the second mixture material
    and :math:`f` is the volume fraction of material a in the material b.

    References:
        * Ph.J. Rouseel; J. Vanhellemont; H.E. Maes. (1993) Thin Solid Films, 234, 423-427
    """

    def get_tensor_fraction(self, lbda: npt.ArrayLike, fraction: float) -> npt.NDArray:
        """Gets the permittivity tensor of the material for wavelength 'lbda',
        while overwriting the set fraction.

        Args:
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).
            fraction (float): Fraction of the guest material used for evaluation. (Range 0 - 1).

        Returns:
            npt.NDArray: Permittivity tensor.
        """
        e_h = self.host_material.get_tensor(lbda)
        e_g = self.guest_material.get_tensor(lbda)
        f = fraction

        mask_equal = np.nonzero(np.equal(e_h, e_g))
        mask_different = np.nonzero(np.not_equal(e_h, e_g))

        p = sqrt(e_h[mask_different]) / sqrt(e_g[mask_different])
        b = 0.25 * ((3 * f - 1) * (1 / p - p) + p)
        z = b + sqrt(power(b, 2) + 0.5)

        e_mix = np.full_like(e_h, np.nan)
        e_mix[mask_equal] = e_h[mask_equal]
        e_mix[mask_different] = (
            z * sqrt(e_h[mask_different]) * sqrt(e_g[mask_different])
        )

        return e_mix



# Encoding: utf-8
"""Sellmeier dispersion."""

import numpy.typing as npt

from .base_dispersion import Dispersion


class SellmeierCustomExponent(Dispersion):
    r"""Sellmeier dispersion with custom exponents.

    Single parameters:
        --

    Repeated parameters:
        :A: Coefficient for n\ :sup:`2` contribution. Defaults to 0.
        :e_A: Exponent for the wavelength in the numerator. Defaults to 1.
        :B: Resonance wavelength. Defaults to 0. Unit in µm\ :sup:`-2`.
        :e_B: Exponent for B. Defaults to 1.

    Output:
        .. math::
            \varepsilon(\lambda) = \sum_j \boldsymbol{A}_j
            \cdot \lambda^{\boldsymbol{e_A}j} /(\lambda^2 - \boldsymbol{B}_j^{\boldsymbol{e_B}j})

        With :math:`j` as the index of the respective oscillator.
    """

    single_params_template = {}
    rep_params_template = {"A": 0, "e_A": 1, "B": 0, "e_B": 1}

    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        lbda = lbda / 1e3
        return sum(
            c.get("A") * lbda ** c.get("e_A") / (lbda**2 - c.get("B") ** c.get("e_B"))
            for c in self.rep_params
        )



# Encoding: utf-8
"""After a calculation is completed, the Solver returns an object of the Result class.

It is used to store the evaluated experiment and all resulting optical properties like
Jones matrices and ellipsometric parameters (psi, delta, rho, mueller matrices).
A list of all properties is given below.

All properties will return an array in the length of the provided wavelength array
of the requested property.

These can be accessed by different methods:

* With dot notation: ``result.property``
* With the get method: ``result.get('property')``

For Matrix properties, a specific value can be requested using a ``property_ij`` notation.
As i and j the respective polarization identifiers or
numerical indices can be used (r/s or R/L or 1...4).

To make handling multiple experiments easier, they can be grouped into a list and provided
to a ResultList object. It provides the same methods for data output as the single Result.
The Output is returned as array over the list of results. If needed, these arrays can be
averaged and used like a Result object for fitting.
"""

from typing import List

import numpy as np
import numpy.typing as npt
from numpy.lib.scimath import sqrt


def _convert_index(index: str) -> int:
    """Return index for character 'index'.

    Args:
        index (str): Polarization index, valid are: 'p', 's', 'R', 'L' or numerical indices '1' till '4'.
    Returns:
        int: 'p', 'L' -> 0
                's', 'R' -> 1
    """
    if index in ["1", "2", "3", "4"]:
        return int(index) - 1
    if index in ["p", "L"]:
        return 0
    if index in ["s", "R"]:
        return 1

    raise ValueError("Wrong index given for variable.")


class Result:
    """Record of a simulation result."""

    @property
    def rho(self) -> npt.NDArray:
        r"""Returns the ellipsometric parameter :math:`\rho` in reflection direction.

        It is calculated by dot product of the \rho matrix
        :math:`M_\rho` and the Jones vector :math:`\vec{E}` of the incident light beam.
        It then takes the :math:`\rho_\text{pp}` element and returns it.

        .. math::
            M_{\text{$\rho$, exp}} = M_\rho \cdot \vec{E}
        """
        rho = np.dot(self.rho_matrix, self.experiment.jones_vector)
        rho = rho[:, 0] / rho[:, 1]

        if self._delta_range == (0, 180):
            rho.imag = -abs(rho.imag)
        return rho

    @property
    def rho_t(self) -> npt.NDArray:
        r"""Returns the ellipsometric parameter :math:`\rho_\text{t}` in transmission direction."""
        rho_t = np.dot(self.rho_matrix_t, self.experiment.jones_vector)
        rho_t = rho_t[:, 0] / rho_t[:, 1]
        if self._delta_range == (0, 180):
            rho_t.imag = -abs(rho_t.imag)
        return rho_t

    @property
    def psi(self) -> npt.NDArray:
        r"""Returns the ellipsometric angle :math:`\psi` in reflection direction.

        It results from:

        .. math::
            \rho = \tan \psi \exp(-i \Delta)
        """
        return np.rad2deg(np.arctan(np.abs(self.rho)))

    @property
    def psi_t(self) -> npt.NDArray:
        r"""Returns the ellipsometric angle :math:`\psi_\text{t}` in transmission direction.

        It results from:

        .. math::
            \rho_\text{t} = \tan \psi_\text{t} \exp(-i \Delta_\text{t})
        """
        return np.rad2deg(np.arctan(np.abs(self.rho_t)))

    @property
    def delta(self) -> npt.NDArray:
        r"""Returns the ellipsometric angle :math:`\Delta` in reflection direction.

        It results from:

        .. math::
            \rho = \tan \psi \exp(-i \Delta)
        """
        if self._delta_range == (0, 360):
            return np.mod(-np.angle(self.rho, deg=True), 360)
        return -np.angle(self.rho, deg=True)

    @property
    def delta_t(self) -> npt.NDArray:
        r"""Returns the ellipsometric angle :math:`\Delta_\text{t}` in transmission direction.

        It results from:

        .. math::
            \rho_\text{t} = \tan \psi_\text{t} \exp(-i \Delta_\text{t})
        """
        if self._delta_range == (0, 360):
            return np.mod(-np.angle(self.rho_t, deg=True), 360)
        return -np.angle(self.rho_t, deg=True)

    @property
    def rho_matrix(self) -> npt.NDArray:
        r"""Returns the matrix of the ellipsometric parameter
        :math:`\rho` in reflection direction.

        .. math::
            M_\rho = \begin{bmatrix}
            \rho_\text{pp} & \rho_\text{ps} \\ \rho_\text{sp} & 1
            \end{bmatrix}
            = r_\text{ss} \begin{bmatrix}
            r_\text{pp}/r_\text{ss} & r_{ps}/r_\text{ss} \\ r_\text{sp}/r_\text{ss} & 1
            \end{bmatrix}
        """
        r_ss = self.jones_matrix_r[..., 1, 1]
        return self.jones_matrix_r / r_ss[:, None, None]

    @property
    def rho_matrix_t(self) -> npt.NDArray:
        r"""Returns the matrix of the ellipsometric parameter
        :math:`\rho_t` in reflection direction.

        .. math::
            M_{\rho \text{,t}} = \begin{bmatrix}
            \rho_\text{t,pp} & \rho_\text{t,ps} \\ \rho_\text{t,sp} & 1
            \end{bmatrix}
            = t_\text{ss} \begin{bmatrix}
            t_\text{pp}/t_\text{ss} & t_\text{ps}/t_\text{ss} \\ t_\text{sp}/t_\text{ss} & 1
            \end{bmatrix}
        """
        t_ss = self.jones_matrix_t[..., 1, 1]
        return self.jones_matrix_t / t_ss[:, None, None]

    @property
    def psi_matrix(self) -> npt.NDArray:
        r"""Returns the matrix of the ellipsometric parameter
        :math:`\psi` in reflection direction.

        .. math::
            M_\psi = \begin{bmatrix}
            \psi_\text{pp} & \psi_\text{ps} \\ \psi_\text{sp} & 45°
            \end{bmatrix}
        """
        return np.rad2deg(np.arctan(np.abs(self.rho_matrix)))

    @property
    def psi_matrix_t(self) -> npt.NDArray:
        r"""Returns the matrix of the ellipsometric parameter
        :math:`\psi_\text{t}` in transmission direction.

        .. math::
            M_{\psi \text{,t}} = \begin{bmatrix}
            \psi_\text{t,pp} & \psi_\text{t,ps} \\ \psi_\text{t,sp} & 45°
            \end{bmatrix}
        """
        return np.rad2deg(np.arctan(np.abs(self.rho_matrix_t)))

    @property
    def delta_matrix(self) -> npt.NDArray:
        r"""Returns the matrix of the ellipsometric parameter
        :math:`\Delta` in reflection direction.

        .. math::
            M_\Delta = \begin{bmatrix}
            \Delta_\text{pp} & \Delta_\text{ps} \\ \Delta_\text{sp} & 0°
            \end{bmatrix}
        """
        return -np.angle(self.rho_matrix, deg=True)

    @property
    def delta_matrix_t(self) -> npt.NDArray:
        r"""Returns the matrix of the ellipsometric parameter
        :math:`\Delta_\text{t}` in transmission direction.

        .. math::
            M_{\Delta \text{,t}} = \begin{bmatrix}
            \Delta_\text{t,pp} & \Delta_\text{t,ps} \\ \Delta_\text{t,sp} & 0°
            \end{bmatrix}
        """
        return -np.angle(self.rho_matrix_t, deg=True)

    @property
    def mueller_matrix(self) -> npt.NDArray:
        """Returns the Mueller matrix for reflection, calculated from the rho matrix."""
        a = np.array([[1, 0, 0, 1], [1, 0, 0, -1], [0, 1, 1, 0], [0, 1j, -1j, 0]])

        # Kronecker product of S and S*
        s_kron_s_star = np.einsum(
            "aij,akl->aikjl", np.conjugate(self.rho_matrix), self.rho_matrix
        ).reshape([self.rho_matrix.shape[0], 4, 4])

        mueller_matrix = np.real(a @ s_kron_s_star @ np.linalg.inv(a))
        mm11 = mueller_matrix[:, 0, 0]

        return mueller_matrix / mm11[:, None, None]

    @property
    def jones_matrix_r(self) -> npt.NDArray:
        r"""Returns the Jones matrix with the amplitude reflection coefficients.

        .. math::
            M_\text{r} = \begin{bmatrix}
            r_\text{pp} & r_\text{ps} \\ r_\text{sp} & r_\text{ss}
            \end{bmatrix}
        """
        return self._jones_matrix_r

    @property
    def jones_matrix_t(self) -> npt.NDArray:
        r"""Returns the Jones matrix with the amplitude transmission coefficients.

        .. math::
            M_\text{t} = \begin{bmatrix}
            t_\text{pp} & t_\text{ps} \\ t_\text{sp} & t_\text{ss}
            \end{bmatrix}
        """
        return self._jones_matrix_t

    @property
    def jones_matrix_rc(self) -> npt.NDArray:
        r"""Returns the Jones matrix with the amplitude reflection coefficients
        for circular polarization.

        .. math::
            M_\text{rc} = \begin{bmatrix}
            r_\text{LL} & r_\text{LR} \\ r_\text{RL} & r_\text{RR}
            \end{bmatrix}
        """
        c = 1 / sqrt(2) * np.array([[1, 1], [1j, -1j]])
        d = 1 / sqrt(2) * np.array([[-1, -1], [-1j, 1j]])
        return np.einsum(
            "ij,...jk,kl->...il", np.linalg.inv(d), self._jones_matrix_r, c
        )

    @property
    def jones_matrix_tc(self) -> npt.NDArray:
        r"""Returns the Jones matrix with the amplitude transmission coefficients
        for circular polarization.

        .. math::
            M_\text{tc} = \begin{bmatrix}
            t_\text{LL} & t_\text{LR} \\ t_\text{RL} & t_\text{RR}
            \end{bmatrix}
        """
        c = 1 / sqrt(2) * np.array([[1, 1], [1j, -1j]])
        return np.einsum(
            "ij,...jk,kl->...il", np.linalg.inv(c), self._jones_matrix_t, c
        )

    @property
    def R(self) -> npt.NDArray:
        r"""Returns the absolute reflectance for unpolarized light.

        .. math::
            R = (R_{pp} + R_{ss}) / 2
        """
        return (self.R_matrix[:, 0, 0] + self.R_matrix[:, 1, 1]) / 2

    @property
    def R_matrix(self) -> npt.NDArray:
        r"""Returns the reflectance matrix separated for s and p polarization.

        .. math::
            M_R = \begin{bmatrix} R_{pp} & R_{ps} \\ R_{sp} & R_{ss} \end{bmatrix}
        """
        return np.abs(self._jones_matrix_r) ** 2

    @property
    def T(self) -> npt.NDArray:
        r"""Returns the absolute transmittance for unpolarized light.

        .. math::
            T = (T_{pp} / T_{ss}) / 2
        """
        return (self.T_matrix[:, 0, 0] + self.T_matrix[:, 1, 1]) / 2

    @property
    def T_matrix(self) -> npt.NDArray:
        r"""Returns the transmittance matrix separated for s and p polarization.

        .. math::
            M_T = \begin{bmatrix} T_{pp} & T_{ps} \\ T_{sp} & T_{ss} \end{bmatrix}
        """
        return np.abs(self._jones_matrix_t) ** 2 * self._power_correction[:, None, None]

    @property
    def Rc_matrix(self) -> npt.NDArray:
        r"""Returns the reflectance matrix for circular polarizations.

        .. math::
            M_{Rc} = \begin{bmatrix} R_{LL} & R_{LR} \\ R_{RL} & R_{RR} \end{bmatrix}
        """
        return np.abs(self.jones_matrix_rc) ** 2

    @property
    def Tc_matrix(self) -> npt.NDArray:
        r"""Returns the transmittance matrix with the for circular polarizations.

        .. math::
            M_{Tc} = \begin{bmatrix} T_{LL} & T_{LR} \\ T_{RL} & T_{RR} \end{bmatrix}
        """
        return np.abs(self.jones_matrix_tc) ** 2 * self._power_correction[:, None, None]

    def __init__(
        self,
        experiment: "Experiment",
        jones_matrix_r: npt.NDArray,
        jones_matrix_t: npt.NDArray,
        power_correction: npt.NDArray = None,
    ) -> None:
        """Creates result object, to store simulation data. Gets called by solvers.

        Args:
            experiment (Experiment):
                Evaluated experiment, with structure and experimental parameters.
            jones_matrix_r (npt.NDArray): Jones matrix for the reflection direction.
            jones_matrix_t (npt.NDArray): Jones matrix for the transmission direction.
            power_correction (npt.NDArray):
                Correction factors, to get the power transmission values.
        """
        self.experiment = experiment
        self._jones_matrix_r = jones_matrix_r
        self._jones_matrix_t = jones_matrix_t
        self._delta_range = (-180, 180)
        if power_correction is None:
            self._power_correction = np.ones_like(self.experiment.lbda)
        else:
            self._power_correction = power_correction

    def get(self, name: str) -> npt.NDArray:
        """Return the data for the requested variable 'name'.

        Args:
            name (str): Variable name to return.
                Examples for 'name':

                * 'r_sp' : Amplitude reflection coefficient from 's' to 'p' polarization.
                * 'r_LR' : Reflection from circular right to circular left polarization.
                * 'T_pp' : Power transmission coefficient from 'p' to 'p' polarization.
                * 'Ψ_ps', 'Δ_pp' : Ellipsometry parameters.
                * 'psi', 'delta', 'rho': Reduced ellipsometry parameters,
                  the whole matrices are returned by 'psi_matrix'.

        Returns:
            npt.NDArray: Array of data.
        """
        return self[name]

    def __getattr__(self, name: str) -> npt.NDArray:
        """Return the data for the requested variable 'name'.

        Args:
            name (str): Variable name to return.
                Examples for 'name'...
                'r_sp' : Amplitude reflection coefficient from 's' to 'p' polarization.
                'r_LR' : Reflection from circular right to circular left polarization.
                'T_pp' : Power transmission coefficient from 'p' to 'p' polarization.
                'Ψ_ps', 'Δ_pp' : Ellipsometry parameters.
                'psi', 'delta', 'rho':
                    Reduced ellipsometry parameters,
                    the whole matrices are returned by 'psi_matrix'.

        Returns:
            npt.NDArray: Array of data.
        """
        names = name.rsplit("_", 1)

        if names[0] == "Ψ":
            names[0] = "psi"
        elif names[0] == "Δ":
            names[0] = "delta"
        elif names[0] == "ρ":
            names[0] = "rho"

        if not (
            names[0] in ["psi", "delta", "rho", "r", "t", "rc", "tc", "Rc", "Tc"]
            or names[0] in self.__dir__()
        ):
            raise AttributeError(f"'Result' object has no attribute '{name}'")

        if len(names) > 1:
            (i, j) = map(_convert_index, names[1])

        if names[0] in ["psi", "delta", "rho", "R", "T"]:
            if len(names) == 1:
                return self.__getattribute__(names[0])
            return self.__getattribute__(names[0] + "_matrix")[:, i, j]

        if names[0] in ["r", "rc", "t", "tc"]:
            if len(names) == 1:
                return self.__getattribute__("jones_matrix_" + names[0])
            return self.__getattribute__("jones_matrix_" + names[0])[:, i, j]

        if names[0] in ["Rc", "Tc"]:
            if len(names) == 1:
                return self.__getattribute__(names[0] + "_matrix")
            return self.__getattribute__(names[0] + "_matrix")[:, i, j]

        return self.__getattribute__(names[0])[:, i, j]

    def as_delta_range(self, lower: int, upper: int):
        """Returns this result in another delta range

        Args:
            lower (int): The lower delta range. Should either be -180 or 0.
            upper (int): The upper delta range. Should either be 180 or 360.

        Raises:
            TypeError: Raised when either lower or upper is not an integer.
            ValueError: Range must be (-180, 180), (0, 180) or (0, 360).
                Otherwise a ValueError is raised.
        """
        if not isinstance(lower, int):
            raise TypeError("lower delta range must be an integer")

        if not isinstance(upper, int):
            raise TypeError("upper delta range must be an integer")

        if (lower, upper) not in [(-180, 180), (0, 180), (0, 360)]:
            raise ValueError(f"Invalid delta range ({lower}, {upper})")

        self._delta_range = (lower, upper)

        return self


class ResultList:
    """Class to make a row of Results easier to handle."""

    def __init__(self, results: List[Result] = None) -> None:
        """Creates an ResultList object.

        Args:
            results (List[Result], optional): List of results to store. Defaults to None.
            mean (bool, optional): Returns the average of all results. Defaults to False.
        """
        if results is None:
            self.results = []
        else:
            self.results = results

    def append(self, result: Result) -> None:
        """Append a single Result to the ResultList.

        Args:
            result (Result): Additional Result to store.
        """
        self.results.append(result)

    def __len__(self) -> int:
        """Returns length of ResultList.

        Returns:
            int: Number of Results in ResultList.
        """
        return len(self.results)

    def __getattr__(self, name: str) -> npt.NDArray:
        """Returns the data for the requested variable 'name' of all results.

        Args:
            name (str): Variable name to return.
                Examples for 'name'...
                'r_sp' : Amplitude reflection coefficient from 's' to 'p' polarization.
                'r_LR' : Reflection from circular right to circular left polarization.
                'T_pp' : Power transmission coefficient from 'p' to 'p' polarization.
                'Ψ_ps', 'Δ_pp' : Ellipsometry parameters.
                'psi', 'delta', 'rho':
                    Reduced ellipsometry parameters,
                    the whole matrices are returned by 'psi_matrix'.
                'mean.name' : Returns averaged values for the requested variable,
                    e.g. 'mean.psi'.

        Returns:
            npt.NDArray: Array of data.
        """

        if name == "mean":
            return AveragedResultList(self.results)

        return np.squeeze(np.array([getattr(result, name) for result in self.results]))


class AveragedResultList(ResultList):
    """ResultList with averaging over all results.
    Can be used as drop-in replacement for Result objects, if for example
    thickness inhomogeneities need to be simulated.
    """

    def __getattr__(self, name: str) -> npt.NDArray:
        """Returns the data for the requested variable 'name' of all results.

        Args:
            name (str): Variable name to return.
                Examples for 'name'...
                'r_sp' : Amplitude reflection coefficient from 's' to 'p' polarization.
                'r_LR' : Reflection from circular right to circular left polarization.
                'T_pp' : Power transmission coefficient from 'p' to 'p' polarization.
                'Ψ_ps', 'Δ_pp' : Ellipsometry parameters.
                'psi', 'delta', 'rho':
                    Reduced ellipsometry parameters,
                    the whole matrices are returned by 'psi_matrix'.

        Returns:
            npt.NDArray: Array of data.
        """

        if name == "mean":
            raise ValueError(
                "The ResultList is already averaged and can't be averaged again."
            )

        return np.mean(super().__getattr__(name), axis=0)



#!/usr/bin/python3
# encoding: utf-8
# %% [markdown]
# # We reproduce results from section 6.4.2 in 'Spectroscopic Ellipsometry', by H. Fujiwara
# Author: O. Castany, M. Müller
#
# Verification of the code against results presented in Fujiwara's book
# p. 241 (section 6.4.2). References in the comments refer to this book.
# Expected results are indicated after the commands. They are identical
# with Fujiwara's results, except for a sign convention for Erp.

# %%
import numpy as np
import elli
from scipy.constants import pi
import matplotlib.pyplot as pyplot
import scipy.linalg

# reduces the number of printed figures in numbers:
np.set_printoptions(suppress=True, precision=4)

# %% [markdown]
# ## Example: Air / anisotropic film / silicon substrate

# %%
n_i = 1.0  # incident medium is air
n_o = 2.0  # ordinary index of thin layer
n_e = 2.5  # extraordinary index of thin layer
lbda = 620  # Wavelength for evaluation (nm)
Phi_i = 70  # 70° indicence angle (degree)
d = 100  # thin layer thickness (nm)
# Orientation of the anisotropy of the thin layer
Phi_E = 45  # 1st Euler angle
Theta_E = 45  # 2nd Eulet angle

# %% [markdown]
# ## Permittivity tensor of the anisotropic film (eq 6.63, p. 241):

# %%
filmMaterial = elli.UniaxialMaterial(
    elli.ConstantRefractiveIndex(n_o), elli.ConstantRefractiveIndex(n_e)
)
R = elli.rotation_euler(Phi_E, Theta_E, 0)
filmMaterial.set_rotation(R)
print(filmMaterial.get_tensor(lbda))

# %% [markdown]
# Should yield:
# ```python
# [[[ 4.5625, -0.5625,  0.7955],
#   [-0.5625,  4.5625, -0.7955],
#   [ 0.7955, -0.7955,  5.125 ]]]
# ```

# %%
air = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n_i))
Kx = n_i * np.sin(np.deg2rad(Phi_i))

print("Kx: {:.4f}".format(Kx))

# %% [markdown]
# ```python
# Kx = 0.9397
# ```

# %%
film = elli.Layer(filmMaterial, d)
epsilon = filmMaterial.get_tensor(lbda)
Delta = elli.Solver4x4.build_delta_matrix(Kx, epsilon)
print("\nDelta matrix:")
print(Delta)

q, Psi = scipy.linalg.eig(Delta[0])
print("\nEigenvalues of the Delta matrix (eq 6.64, p.241):")
print(np.real(q))

# %% [markdown]
# ```python
# # Delta matrix:
# [[[-0.1459,  0.1459,  0.    ,  0.8277],
#   [ 0.    ,  0.    , -1.    ,  0.    ],
#   [ 0.439 , -3.556 ,  0.    , -0.1459],
#   [ 4.439 , -0.439 ,  0.    , -0.1459]]]
#
# # Eigenvalues of Delta
# [-2.174 , -1.7655,  1.7655,  1.8822]
# ```

# %%
Tp = elli.PropagatorExpm().calculate_propagation(Delta, -d, np.array([lbda]))
print("\nPropagation matrix (eq 6.66, p. 242):")
print(Tp)

# %% [markdown]
# ```python
# [[[-0.353-0.057j,  0.091-0.001j,  0.033+0.044j,  0.056-0.397j],
#   [ 0.104+0.08j , -0.327-0.011j,  0.004+0.502j, -0.033-0.044j],
#   [ 0.162-0.014j, -0.017+1.753j, -0.327-0.011j, -0.091+0.001j],
#   [ 0.300-2.12j , -0.162+0.014j, -0.104-0.08j , -0.353-0.057j]]]
# ```

# %%
n_t = 3.898 + 0.016j  #  refractive index of substrate
silicon = elli.ConstantRefractiveIndex(n_t).get_mat()
s = elli.Structure(air, [film], silicon)
result = s.evaluate(np.array([lbda]), Phi_i)
# T = s.getStructureMatrix(Kx, k0)
# print("\nTransfer matrix T (eq 6.67, p. 242):")
# print(T)
"""                                                                 (eq 6.67)
Fujiwara uses the ellipsometry convention for the orientation of the 'p'
polarized electric fields (see figures 2.15 and 6.14). With my convention,
Erp is reversed. Consequently, T_{4j} and T_{i4} have a reversed sign, except
T_{44}.
T =
matrix([[-1.949-3.588j,  1.671-1.548j,  0.273-0.03j , -0.630+0.147j],
        [ 1.617+1.679j, -1.992+3.434j, -0.301-0.064j,  0.861+0.101j],
        [ 0.065-0.086j,  0.039+0.097j, -0.712-3.485j,  0.004+1.265j],
        [-0.167-0.403j,  0.593+0.386j,  0.370-1.199j, -1.662+3.093j]])
"""
Jr = result.jones_matrix_r  #  Jones reflexion matrix
print("\nJones reflexion matrix (p. 243):")
print(Jr)

print("\nEllipsometry parameters (p. 243):")
(Psi, Delta) = (result.psi_matrix, result.delta_matrix)
print("Psi\n" + str(Psi))
print("Delta\n" + str(Delta))

# %% [markdown]
# ```python
# Jr =
# [[[ 0.310+0.161j,  0.107+0.002j],
#   [ 0.042-0.036j, -0.552+0.151j]]]
# #The first line, [r_pp, r_ps], has reversed sign compared to Fujiwara's result, due to the different convention for Erp.
# Ellipsometry parameters:
# Psi                         Delta
# [[ 31.4419  10.5641]         [[ -42.734   -16.4123]
#  [  5.5332  45.    ]]         [-155.024     0.    ]]
# ```

# %%
# We now reproduce figure 6.19:
print("\nWe reproduce figure 6.19, p. 242...")

# %%
Theta_E_list = [0, 45, 90]
Phi_E_list = np.linspace(0, 180, 36 * 2 + 1)

data = elli.ResultList()

for Theta_E in Theta_E_list:
    for Phi_E in Phi_E_list:
        R = elli.rotation_euler(Phi_E, Theta_E, 0)
        filmMaterial.set_rotation(R)
        data.append(s.evaluate(np.array([lbda]), Phi_i))

# %%
psi_pp = data.psi_pp.reshape(3, 73).T
psi_ps = data.psi_ps.reshape(3, 73).T

delta_pp = data.delta_pp.reshape(3, 73).T
delta_ps = data.delta_ps.reshape(3, 73).T

# %%
fig = pyplot.figure()
# Plot curves for the three values of Theta_E
pyplot.rcParams["axes.prop_cycle"] = pyplot.cycler("color", "kbg")
ax1 = fig.add_subplot(2, 2, 1)
ax1.plot(Phi_E_list, psi_pp)
ax1.set_ylabel(r"$\Psi_{pp}$")

ax2 = fig.add_subplot(2, 2, 2)
ax2.plot(Phi_E_list, delta_pp)
ax2.set_ylabel(r"$\Delta_{pp}$")

# Plot curves for two values of Theta_E
pyplot.rcParams["axes.prop_cycle"] = pyplot.cycler("color", "kbg")
ax3 = fig.add_subplot(2, 2, 3)
ax3.plot(Phi_E_list, psi_ps)
ax3.set_ylabel(r"$\Psi_{ps}$")
ax3.set_xlabel(r"$\phi_E$")

ax4 = fig.add_subplot(2, 2, 4)
ax4.plot(Phi_E_list, delta_ps)
ax4.set_ylabel(r"$\Delta_{ps}$")
ax4.set_xlabel(r"$\phi_E$")

pyplot.tight_layout()
pyplot.show()

# %%



"""Tests the importing from the refractiveindex.info database."""

import elli
import numpy as np
import pytest


class TestRefractiveIndexInfo:
    """Test if the refractive index.info parser works."""

    RII = elli.db.RII()

    def test_get_mat(self):
        mat = self.RII.get_mat("Au", "Johnson")
        assert isinstance(mat, elli.IsotropicMaterial)

    def test_dispersion_error(self):
        with pytest.raises(ValueError):
            self.RII.get_dispersion("foo", "bar")

    def test_tabular_nk(self):
        disp = self.RII.get_dispersion("Au", "Johnson")  # tabular nk

        np.testing.assert_almost_equal(
            disp.get_refractive_index(310.7), 1.53 + 1j * 1.893, decimal=3
        )
        np.testing.assert_almost_equal(
            disp.get_refractive_index(1088), 0.27 + 1j * 7.150, decimal=3
        )

    def test_tabular_n(self):
        disp = self.RII.get_dispersion("Xe", "Koch")  # Tabular n

        np.testing.assert_almost_equal(
            disp.get_refractive_index(234.555), 1.00084664, decimal=6
        )
        np.testing.assert_almost_equal(
            disp.get_refractive_index(612.327), 1.00070157, decimal=6
        )

    def test_formula_1(self):
        disp = self.RII.get_dispersion("SrTiO3", "Dodge")  # Formula 1

        np.testing.assert_almost_equal(
            disp.get_refractive_index(500), 2.4743, decimal=4
        )
        np.testing.assert_almost_equal(
            disp.get_refractive_index(1000), 2.3160, decimal=4
        )

    def test_formula_2_tabular_k(self):
        disp = self.RII.get_dispersion("SCHOTT-BK", "N-BK7")  # Formula 2 + k

        np.testing.assert_almost_equal(
            disp.get_refractive_index(500), 1.5214 + 1j * 9.5781e-9, decimal=4
        )
        np.testing.assert_almost_equal(
            disp.get_refractive_index(1000), 1.5075 + 1j * 9.9359e-9, decimal=4
        )

    def test_formula_3(self):
        disp = self.RII.get_dispersion("HOYA-NbF", "NBF1")  # Formula 3

        np.testing.assert_almost_equal(
            disp.get_refractive_index(500), 1.7520 + 1j * 3.9809e-9, decimal=4
        )
        np.testing.assert_almost_equal(
            disp.get_refractive_index(1000), 1.7271 + 1j * 7.9617e-9, decimal=4
        )

    def test_formula_4(self):
        disp = self.RII.get_dispersion("AgCl", "Tilton")  # formula 4

        np.testing.assert_almost_equal(
            disp.get_refractive_index(1024), 2.0213900020277, decimal=12
        )
        np.testing.assert_almost_equal(
            disp.get_refractive_index(10080), 1.9799748188746, decimal=12
        )

    def test_formula_5(self):
        disp = self.RII.get_dispersion("SF6", "Vukovic")  # Formula 5

        np.testing.assert_almost_equal(
            disp.get_refractive_index(600), 1.00072905, decimal=6
        )
        np.testing.assert_almost_equal(
            disp.get_refractive_index(1000), 1.00072017, decimal=6
        )

    def test_search(self):
        assert len(self.RII.search("")) == 0
        assert len(self.RII.search("C")) > len(self.RII.search("C", fuzzy=False))
        assert len(self.RII.search("Au", "book")) > len(
            self.RII.search("Au", "book", fuzzy=False)
        )



# Encoding: utf-8
"""Lorentz dispersion law with parameters in units of energy."""

import numpy.typing as npt

from ..utils import conversion_wavelength_energy
from .base_dispersion import Dispersion


class LorentzEnergy(Dispersion):
    r"""Lorentz dispersion law with parameters in units of energy.

    Single parameters:
        --

    Repeated parameters:
        :A: Amplitude of the oscillator. Defaults to 1.
        :E: Resonance energy. Defaults 0. Unit in eV.
        :gamma: Broadening of the oscillator. Defaults to 0. Unit in eV.

    Output:
        .. math::
            \varepsilon(E) = 1 + \sum_j \boldsymbol{A}_j / \left(E^2-\boldsymbol{E}_j^2
            + i \cdot \boldsymbol{gamma}_j \cdot E\right)

        With :math:`j`  as the index for the respective oscillator.
    """

    single_params_template = {}
    rep_params_template = {"A": 1, "E": 0, "gamma": 0}

    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        energy = conversion_wavelength_energy(lbda)
        return 1 + sum(
            c.get("A") / (c.get("E") ** 2 - energy**2 - 1j * c.get("gamma") * energy)
            for c in self.rep_params
        )



# Encoding: utf-8
from abc import ABC, abstractmethod
from copy import deepcopy

from .result import Result


class Solver(ABC):
    """
    Solver base class to evaluate Experiment objects.
    Here the experiment and structure get unpacked
    and the simulation results get returned.

    The actual simulation is handled by subclasses.
    Therefore, this class should never be called directly.
    """

    experiment = None
    structure = None
    lbda = None
    theta_i = None
    jones_vector = None
    permittivity_profile = None

    @abstractmethod
    def calculate(self) -> Result:
        pass

    def __init__(self, experiment: "Experiment") -> None:
        self.experiment = deepcopy(experiment)
        self.structure = self.experiment.structure
        self.lbda = self.experiment.lbda
        self.theta_i = self.experiment.theta_i
        self.jones_vector = self.experiment.jones_vector
        self.permittivity_profile = self.structure.get_permittivity_profile(self.lbda)



"""Tests for the materials classes"""

import elli
from pytest import raises


class TestMaterials:
    disp = elli.ConstantRefractiveIndex(2)
    disp2 = elli.ConstantRefractiveIndex(5)
    mat = elli.IsotropicMaterial(disp)
    mat2 = elli.IsotropicMaterial(disp2)

    def test_materials_typeguard(self):
        """Checks if materials check input types"""
        with raises(TypeError):
            elli.IsotropicMaterial(self.mat)

        with raises(TypeError):
            elli.IsotropicMaterial(23)

        with raises(TypeError):
            elli.UniaxialMaterial(self.disp, 23)

        with raises(TypeError):
            elli.BiaxialMaterial(self.disp, self.disp, 23)

        with raises(TypeError):
            elli.VCAMaterial(self.mat, self.disp, 0.5)

        with raises(TypeError):
            elli.VCAMaterial(self.disp, self.mat, 0.5)

        with raises(ValueError):
            elli.VCAMaterial(self.mat, self.mat, 10)



# Encoding: utf-8
import numpy as np
from numpy.lib.scimath import sqrt

try:
    import matplotlib.pyplot as plt
except ImportError as e:
    raise ImportError(
        "Optional dependency matplotlib missing. This module will not work properly.\n"
        "Try installing this package with the additional fitting requirement, "
        "i.e. pip install pyElli[fitting]"
    ) from e

from ..utils import E_X


def get_permittivity_profile(structure, lbda):
    """Returns permittivity tensor profile."""
    layers = []
    for L in structure.layers:
        layers.extend(L.get_permittivity_profile(lbda))
    front = (float("inf"), structure.front_material.get_tensor(lbda))
    back = (float("inf"), structure.back_material.get_tensor(lbda))
    return sum([[front], layers, [back]], [])


def get_index_profile(structure, lbda, v=E_X):
    """Returns refractive index profile.

    'v' : Unit vector, direction of evaluation of the refraction index.
          Default value is v = e_x.
    """
    profile = get_permittivity_profile(structure, lbda)
    (h, epsilon) = list(zip(*profile))  # unzip
    n = [sqrt((v.T * eps * v)[0, 0, 0]) for eps in epsilon]
    return list(zip(h, n))


def draw_structure(structure, lbda=1000, method="graph", margin=0.15):
    """Draw the structure.

    'method' : 'graph' or 'section'
    Returns : Axes object
    """
    # Build index profile
    profile = get_index_profile(structure, lbda)
    (h, n) = list(zip(*profile))  # unzip
    n = np.array(n)
    z_layers = np.hstack((0.0, np.cumsum(h[1:-1])))
    z_max = z_layers[-1]
    if z_max != 0.0:
        z_margin = margin * z_max
    else:
        z_margin = 1e-6
    z = np.hstack((-z_margin, z_layers, z_max + z_margin))
    # Call specialized methods
    if method == "graph":
        ax = _draw_structure_graph(structure, z, n)
    elif method == "section":
        ax = _draw_structure_section(structure, z, n)
    else:
        ax = None
    return ax


def _draw_structure_graph(structure, z, n):
    """Draw a graph of the refractive index profile"""
    n = np.hstack((n, n[-1]))
    # Draw the graph
    fig = plt.figure(figsize=(8, 3))
    ax = fig.add_subplot(1, 1, 1)
    fig.subplots_adjust(bottom=0.17)
    ax.step(z, n.real, "black", where="post")
    ax.spines["top"].set_visible(False)
    ax.xaxis.set_ticks_position("bottom")
    ax.set_xlabel("z (nm)")
    ax.set_ylabel("n'")
    ax.set_xlim(z.min(), z.max())
    ax.set_ylim(bottom=1.0)
    return ax


def _draw_structure_section(structure, z, n):
    """Draw a cross section of the structure"""
    # Prepare arrays for pcolormesh()
    X = z * np.ones((2, 1))
    Y = np.array([0, 1]).reshape((2, 1)) * np.ones_like(z)
    n = np.array(n).reshape((1, -1)).real
    # Draw the cross section
    fig = plt.figure(figsize=(8, 3))
    ax = fig.add_subplot(1, 1, 1)
    fig.subplots_adjust(left=0.05, bottom=0.15)
    ax.set_yticks([])
    ax.set_xlabel("z (nm)")
    ax.set_xlim(z.min(), z.max())
    stack = ax.pcolormesh(X, Y, n, cmap=plt.get_cmap("gray_r"))
    colbar = fig.colorbar(
        stack, orientation="vertical", anchor=(1.2, 0.5), fraction=0.05
    )
    colbar.ax.set_xlabel("n'", position=(3, 0))
    return ax
